"""
Rule Based System using Nest Simulator
"""

from stateMachineClass import FSAHelperFunctions
import pyNN.nest as sim

fsa = FSAHelperFunctions("nest")

class RBS:

    def __init__(self):
        self.factGroups = {}
        self.rules = {}
        self.retractions = {}
        self.assertions = {}
        self.interns = []

    def findGroup(self, name):
        if(name not in self.factGroups):
            return None
        return self.factGroups[name]

    def getGroup(self, name):
        group = self.findGroup(name)
        if(group == None):
            group = []
            self.factGroups[name] = group
        
        return group

    def addFact(self, fact, active = True):
        group = self.getGroup(fact[0])

        ca = sim.Population(fsa.CA_SIZE, sim.IF_cond_exp, fsa.CELL_PARAMS, label = fact)
        fsa.makeCA(ca, 0)

        if(active):
            spikeTimes = {'spike_times': [[sim.get_current_time()+5]]}
            spikeGen = sim.Population(1, sim.SpikeSourceArray, spikeTimes)
            fsa.turnOnStateFromSpikeSource(spikeGen,ca,0)

        ca.record("spikes")

        f = (fact[1],ca)
        group.append([f])

        self.applyRulesToFacts()

        return f

    def getFact(self, fact):
        group = self.getGroup(fact[0])

        found = None
        for f in group:
            match = True
            for i,p in enumerate(f[0][0]):
                if(p != fact[1][i]):
                    match = False
                    break
            if(match):
                found = f[0]
        
        if(found == None):
            found = self.addFact(fact, active = False)
        
        return found


    def addRule(self, rule):
        self.rules[rule[0]] = rule[1]

        self.applyRulesToFacts()

    def updateFact(self, factToUpdate,  fact):
        ca = sim.Population(fsa.CA_SIZE, sim.IF_cond_exp, fsa.CELL_PARAMS, label = fact)
        fsa.makeCA(ca, 0)
        ca.record("spikes")

        factToUpdate[1].append((fact[1][0],ca))

    def getMatchingFacts(self, positive,name,properties):
        if(name not in self.factGroups):
            return None
        
        group = self.factGroups[name]
        found = []

        for fact in group:
            match = True
            for propGroup in fact:
                for i,prop in enumerate(propGroup[0]):

                    if(match == False):
                        break
                    val = properties[i]

                    #is anything?
                    if(val == "?" and positive):
                        continue

                    if(isinstance(val, str)):
                        lVal = list(val)
                        #is variable?
                        if(lVal[0] == "?"):
                            val = val[:1]

                    if(positive):
                        if(val != "?" and val != prop):
                            match = False
                    else:
                        if(val == "?" or val == prop):
                            match = False
                                
                if(match):
                    found.append(propGroup)
                    
        return found

    def findMatches(self, precon):
        return self.getMatchingFacts(precon[0], precon[1], precon[2])

    def getSearch(self, item, variables):
        params = []            
        for p in item[2]:
            if(p in variables):
                params.append(variables[p])
            else:
                params.append(p)

        search = (item[0],item[1],tuple(params),item[3])

        return search

    def extractVariables(self, var, found, precon):
        for i,prop in enumerate(precon):
            if(len(prop) > 1 and prop[0] == "?" and prop not in var):
                var[prop] = found[0][i]
        return var

    def buildFounds(self, tree, item):
        newTree = []
        for f in tree:
            search = self.getSearch(item, f[0])
            founds = self.findMatches(search)
            if(founds == None):
                continue
            for found in founds:
                var = f[0].copy()
                var = self.extractVariables(var, found, item[2])
                newTree.append((var, f[1] + [(found,item[3])]))

        return newTree

    def extractValue(self, item, match):
        if(isinstance(item, tuple)):
            c = item[0]
            left = item[1]
            right = item[2]
            if(isinstance(left, str) and left[0] == "?"):
                left = match[left]
            
            if(isinstance(right, str) and right[0] == "?"):
                right = match[right]

            if(c == "+"):
                value = left + right
            elif(c == "-"):
                value = left - right
            elif(c == "*"):
                value = left * right
            elif(c == "/"):
                value = left / right
        else:
            if(isinstance(item, str) and item[0] == "?"):
                value = match[item]
            else:
                value = item
        
        return value

    def testVariables(self, match, item):
        op = item[1]
        left = self.extractValue(item[2], match[0])
        right = self.extractValue(item[3], match[0])

        if(op == ">"):
            return left > right
        elif(op == "<"):
            return left < right
        elif(op == "="):
            return left == right
        elif(op == "<>"):
            return left != right
        else:
            return False

    def applyRulesToFacts(self):
        for key in list(self.rules):
            rule = self.rules[key]
            ifs = rule[0]
            thens = rule[1]

            matches = [({},[])]
            for i in range(0, len(ifs)):
                if(ifs[i][0] == "Test"):
                    continue
                else:
                    matches = self.buildFounds(matches, ifs[i])


            validMatches = []
            for m in range(0, len(matches)):
                isMatch = True
                for i in range(0, len(ifs)):
                    if(ifs[i][0] == "Test"):
                        isMatch = self.testVariables(matches[m], ifs[i])
                        if(isMatch == False):
                            break
                    else:
                        continue
                if(isMatch):
                    validMatches.append(matches[m])
                

            for m in validMatches:
                assertions = []
                retractions = []
                for t in thens:
                    option,item = t
                    if(option == "assert"):
                        assertions.append(item)
                        #self.addAssertion(item, m)
                    elif(option == "retract"):
                        retractions.append(item)
                        #self.addRetraction(item, m[1]) 
                    else:
                        raise "Invalid Rule Configuration"

                self.addAssertions(assertions, m)
                self.addRetractions(retractions, m[1])


    def twoTurnOnOne(self,pop1,pop2,pop3):
        #print "[{} and {}] turns on [{}]".format(pop1.label,pop2.label,pop3.label)
        fsa.stateHalfTurnsOnState(pop1,0,pop3,0)
        fsa.stateHalfTurnsOnState(pop2,0,pop3,0)

    def addIntermediate(self, pop1, pop2):
        intermediate = sim.Population(fsa.CA_SIZE, sim.IF_cond_exp, fsa.CELL_PARAMS, label="{} and {}".format(pop1.label,pop2.label))
        intermediate.record("spikes")
        fsa.makeCA(intermediate, 0)
        self.twoTurnOnOne(pop1,pop2,intermediate)
        self.interns.append(intermediate)
        return intermediate

    def matchesTurnOnOperator(self, match, ca):
        if len(match) == 1:
            fsa.stateTurnsOnState(match[0][0][1],0,ca,0)
        elif (len(match) == 2):
            self.twoTurnOnOne(match[0][0][1],match[1][0][1],ca)
        else:
            index = 0

            pop1 = match[index][0][1]
            pop2 = match[index+1][0][1]

            intermediate = self.addIntermediate(pop1,pop2)
            while(index < len(match)-2):

                pop3 = match[index+2][0][1]

                if(index == len(match)-3):
                    self.twoTurnOnOne(intermediate,pop3,ca)
                    fsa.stateTurnsOffState(ca,0,intermediate,0)
                else:
                    inter2 = self.addIntermediate(intermediate,pop3)
                    fsa.stateTurnsOffState(inter2,0,intermediate,0)
                    intermediate = inter2

                index = index + 1

    def addAssertions(self, asssertions, match):
        label = ""
        for i,m in enumerate(match[1]):
            newLabel = "({} == {})".format(m[1],m[0][1].label)
            if (i == 0):
                label = label+newLabel
            else:
                label = label+" and "+newLabel

        if(label in self.assertions):
            return

        facts = []
        for assertion in asssertions:
            variables = match[0]
            properties = assertion[1]
            newProps = []
            for p in properties:
                prop = self.extractValue(p, variables)
                newProps.append(prop)
        
            fact = (assertion[0],tuple(newProps))

            # allocate temporary before adding new fact
            self.assertions[label] = None

            # create new fact to be asserted
            fact = self.getFact(fact)
            facts.append(fact)
        
        #create assertion ca
        aaPop = sim.Population(fsa.CA_SIZE, sim.IF_cond_exp, fsa.CELL_PARAMS, label="{} asserts {}".format(label, fact[1].label))
        aaPop.record("spikes")
        fsa.makeCA(aaPop, 0)
        self.assertions[label] = aaPop
        
        for fact in facts:
        #apply assertion
            fsa.stateTurnsOnState(aaPop,0,fact[1],0)
            fsa.stateTurnsOffState(aaPop,0,aaPop,0)

        self.matchesTurnOnOperator(match[1], aaPop)
    
    def addRetractions(self, retractions, match):
        label = ""
        for i,m in enumerate(match):
            newLabel = "({} == {})".format(m[1],m[0][1].label)
            if (i == 0):
                label = label+newLabel
            else:
                label = label+" and "+newLabel

        if(label in self.retractions):
            return

        #create retraction CA
        rePop = sim.Population(fsa.CA_SIZE, sim.IF_cond_exp, fsa.CELL_PARAMS, label="{} retracts {}".format(label, retractions))
        rePop.record("spikes")
        fsa.makeCA(rePop, 0)
        self.retractions[label] = rePop
        fsa.stateTurnsOffState(rePop,0,rePop,0)
        
        #apply retraction
        for retraction in retractions:
            for m in match: 
                if m[1] == retraction:
                    turnOfCa = m[0][1]       
            fsa.stateTurnsOffState(rePop, 0, turnOfCa, 0)

        self.matchesTurnOnOperator(match, rePop)
    
